/*
 * This Software is licensed under one of the following licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 *
 * 2) under the terms of the "The BSD License" a copy of which is
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * 3) under the terms of the "GNU General Public License (GPL) Version 2" a
 *    copy of which is available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/gpl-license.php.
 *
 * Licensee has the right to choose one of the above licenses.
 *
 * Redistributions of source code must retain the above copyright
 * notice and one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/***************************************************************************
 *
 *   Module:		 uDAPL
 *
 *   Filename:		 dapl_ib_cm.c
 *
 *   Author:		 Arlin Davis
 *
 *   Created:		 3/10/2005
 *
 *   Description: 
 *
 *   The uDAPL openib provider - connection management
 *
 ****************************************************************************
 *		   Source Control System Information
 *
 *    $Id: $
 *
 * Copyright (c) 2005 Voltaire Inc.  All rights reserved.
 * Copyright (c) 2005 Intel Corporation. All rights reserved.
 * Copyright (c) 2004-2005, Mellanox Technologies, Inc. All rights reserved. 
 * Copyright (c) 2003 Topspin Corporation.  All rights reserved. 
 * Copyright (c) 2005 Sun Microsystems, Inc. All rights reserved.
 *
 **************************************************************************/

#include "dapl.h"
#include "dapl_adapter_util.h"
#include "dapl_ep_util.h"
#include "dapl_evd_util.h"
#include "dapl_cr_util.h"
#include "dapl_name_service.h"
#include "dapl_ib_util.h"
#include <sys/poll.h>
#include <signal.h>

/* prototypes */
static void dapli_path_comp_handler(uint64_t req_id, void *context, int rec_num);
static void dapli_rt_comp_handler(uint64_t req_id, void *context, int rec_num);
static void dapli_rep_recv(struct dapl_cm_id *conn, struct ib_cm_event *rep_recv_param);
static struct dapl_cm_id * dapli_req_recv(struct dapl_cm_id *conn, struct ib_cm_event *event);
static int dapli_cm_active_cb(struct dapl_cm_id *conn, struct ib_cm_event *event);
static int dapli_cm_passive_cb(struct dapl_cm_id *conn, struct ib_cm_event *event);

/* prototypes */
#if __BYTE_ORDER == __LITTLE_ENDIAN
static inline uint64_t cpu_to_be64(uint64_t x) { return bswap_64(x); }
#elif __BYTE_ORDER == __BIG_ENDIAN
static inline uint64_t cpu_to_be64(uint64_t x) { return x; }
#endif


void dapli_ip_comp_handler(uint64_t req_id, void *context, int rec_num)
{
	struct dapl_at_record	*at_rec = context;
	struct sockaddr_in      *ipv4_addr = (struct sockaddr_in*)at_rec->addr;
	int			status;

	dapl_dbg_log(DAPL_DBG_TYPE_CM,
		     " ip_comp_handler: rec %p ->id %lld id %lld num %d %x\n",
		     context, at_rec->req_id, req_id, rec_num,
		     ipv4_addr->sin_addr.s_addr);

        if (rec_num <= 0)  {
		struct ib_at_completion at_comp;

                dapl_dbg_log(DAPL_DBG_TYPE_CM,
			     " ip_comp_handler: resolution err %d retry %d\n",
			     rec_num, at_rec->retries + 1);

		ipv4_addr->sin_addr.s_addr = 0;

		if ((++at_rec->retries > IB_MAX_AT_RETRY) || (rec_num == -EINTR))
                        goto bail;

		at_comp.fn = dapli_ip_comp_handler;
		at_comp.context = at_rec;
		
		status = ib_at_ips_by_gid(&at_rec->hca_ptr->ib_trans.gid, 
					  &ipv4_addr->sin_addr.s_addr, 1,
					  &at_comp, &at_rec->req_id);
		if (status < 0) 
			goto bail;

		dapl_dbg_log(DAPL_DBG_TYPE_UTIL,
			" ip_comp_handler: ips_by_gid %d rec %p->id %lld\n",
			status, at_rec, at_rec->req_id );
        } 

	if (ipv4_addr->sin_addr.s_addr)
		dapl_os_wait_object_wakeup(at_rec->wait_object);

	return;
bail:
	dapl_dbg_log(DAPL_DBG_TYPE_CM,
		     " ip_comp_handler: ERR: at_rec %p, id %lld num %d\n",
		     at_rec, req_id, rec_num);

	dapl_os_wait_object_wakeup(at_rec->wait_object);
}

static void dapli_path_comp_handler(uint64_t req_id, void *context, int rec_num)
{
	struct dapl_cm_id *conn = context;
	int status;
	ib_cm_events_t event;

	dapl_dbg_log(DAPL_DBG_TYPE_CM,
		     " path_comp_handler: ctxt %p, req_id %lld rec_num %d\n",
		     context, req_id, rec_num);

	dapl_dbg_log(DAPL_DBG_TYPE_CM, 
		" path_comp_handler: SRC GID subnet %016llx id %016llx\n",
		(unsigned long long)cpu_to_be64(conn->dapl_path.sgid.global.subnet_prefix),
		(unsigned long long)cpu_to_be64(conn->dapl_path.sgid.global.interface_id) );

	dapl_dbg_log(DAPL_DBG_TYPE_CM, 
		" path_comp_handler: DST GID subnet %016llx id %016llx\n",
		(unsigned long long)cpu_to_be64(conn->dapl_path.dgid.global.subnet_prefix),
		(unsigned long long)cpu_to_be64(conn->dapl_path.dgid.global.interface_id) );

	dapl_dbg_log(DAPL_DBG_TYPE_CM, 
		" path_comp_handler: slid %x dlid %x mtu %x(%x) pktlife %x(%x)\n",
		ntohs(conn->dapl_path.slid), ntohs(conn->dapl_path.dlid),
		conn->dapl_path.mtu, conn->dapl_path.mtu_selector,
		conn->dapl_path.packet_life_time, 
		conn->dapl_path.packet_life_time_selector );

	dapl_dbg_log(DAPL_DBG_TYPE_CM, 
		" path_comp_handler: hops %x npaths %x pkey %x tclass %x rate %x(%x)\n",
		conn->dapl_path.hop_limit, conn->dapl_path.numb_path,
		conn->dapl_path.pkey, conn->dapl_path.traffic_class,
		conn->dapl_path.rate, conn->dapl_path.rate_selector);
	
	if (rec_num <= 0) {
		dapl_dbg_log(DAPL_DBG_TYPE_CM, 
			     " path_comp_handler: ERR %d retry %d\n",
			     rec_num, conn->retries + 1);
		if (++conn->retries > IB_MAX_AT_RETRY) {
			dapl_dbg_log(DAPL_DBG_TYPE_CM,
				" path_comp_handler: ERR no PATH (ep=%p)\n",
				conn->ep);
			event = IB_CME_DESTINATION_UNREACHABLE;
			goto bail;
		}

		status = ib_at_paths_by_route(&conn->dapl_rt, 0,
					      &conn->dapl_path, 1,
					      &conn->dapl_comp, 
					      &conn->dapl_comp.req_id);
		if (status) {
			dapl_dbg_log(DAPL_DBG_TYPE_ERR,
				     " path_by_route: retry ERR %d id %lld\n",
				     status, conn->dapl_comp.req_id);
			event = IB_CME_LOCAL_FAILURE;
			goto bail;
		}
		return;
	}

	/* Mellanox performance workaround - best performance is MTU of 1024 */
	if (conn->dapl_path.mtu > IBV_MTU_1024)    
		conn->dapl_path.mtu = IBV_MTU_1024;

	conn->req.service_id = cpu_to_be64(conn->service_id);
	conn->req.primary_path = &conn->dapl_path;
	conn->req.alternate_path = NULL;

	status = ib_cm_send_req(conn->cm_id, &conn->req);
	if (status) {
		dapl_dbg_log(DAPL_DBG_TYPE_ERR, " ib_send_cm_req failed: %s\n",
			     strerror(errno));
		event = IB_CME_LOCAL_FAILURE;
		goto bail;
	}
	return;

bail:
	dapl_evd_connection_callback(conn, event, NULL, conn->ep);
}

static void dapli_rt_comp_handler(uint64_t req_id, void *context, int rec_num)
{
	struct dapl_cm_id *conn = context;
	int status;
	ib_cm_events_t event;

	dapl_dbg_log(DAPL_DBG_TYPE_CM,
		     " rt_comp_handler: conn %p, req_id %lld rec_num %d\n",
		     conn, req_id, rec_num);

	if (rec_num <= 0) {
		dapl_dbg_log(DAPL_DBG_TYPE_ERR,
			     " dapl_rt_comp_handler: ERROR rec %d retry %d\n",
			     rec_num, conn->retries+1 ); 

		if (++conn->retries > IB_MAX_AT_RETRY) {
			event = IB_CME_DESTINATION_UNREACHABLE;
			goto bail;
		}

		status = ib_at_route_by_ip(
			((struct sockaddr_in *)&conn->r_addr)->sin_addr.s_addr,
			0, 0, IB_AT_ROUTE_FORCE_ATS, 
			&conn->dapl_rt, 
			&conn->dapl_comp,&conn->dapl_comp.req_id);
		if (status < 0) {
			dapl_dbg_log(DAPL_DBG_TYPE_ERR, "dapl_rt_comp_handler: "
				    "ib_at_route_by_ip failed with status %d\n",
				    status);
			event = IB_CME_DESTINATION_UNREACHABLE;
			goto bail;
		}
		if (status == 1)
			dapli_rt_comp_handler(conn->dapl_comp.req_id, conn, 1);

		return;
	}

	if (!conn->dapl_rt.dgid.global.subnet_prefix || 
		req_id != conn->dapl_comp.req_id) {
		dapl_dbg_log(DAPL_DBG_TYPE_ERR,
			     " dapl_rt_comp_handler: ERROR: cb id=%d(%d)\n",
			     req_id, conn->dapl_comp.req_id ); 
		return;
	}

	conn->dapl_comp.fn = &dapli_path_comp_handler;
	conn->dapl_comp.context = conn;
	conn->retries = 0;
	status = ib_at_paths_by_route(&conn->dapl_rt, 0, &conn->dapl_path, 1,
				      &conn->dapl_comp, 
				      &conn->dapl_comp.req_id);
	if (status) {
		dapl_dbg_log(DAPL_DBG_TYPE_ERR,
			     "dapl_rt_comp_handler: ib_at_paths_by_route "
			     "returned %d id %lld\n", status, 
			     conn->dapl_comp.req_id);
		event = IB_CME_LOCAL_FAILURE;
		goto bail;
	}
	return;

bail:
	dapl_evd_connection_callback(conn, event, NULL, conn->ep);
}

static void dapli_destroy_cm_id(struct dapl_cm_id *conn)
{
	int in_callback;

	dapl_dbg_log(DAPL_DBG_TYPE_CM, 
		     " destroy_cm_id: conn %p id %d\n",conn,conn->cm_id);

	dapl_os_lock(&conn->lock);
	conn->destroy = 1;
	in_callback = conn->in_callback;
	dapl_os_unlock(&conn->lock);

	if (!in_callback) {
		ib_cm_destroy_id(conn->cm_id);
		if (conn->ep)
			conn->ep->cm_handle = IB_INVALID_HANDLE;

		dapl_os_free(conn, sizeof(*conn));
	}
}

static void dapli_rep_recv(struct dapl_cm_id  *conn,
			   struct ib_cm_event *event)
{
	int status;

	if (conn->ep->qp_handle == IB_INVALID_HANDLE) {
		dapl_dbg_log(DAPL_DBG_TYPE_ERR, 
			     " dapli_rep_recv: invalid qp "
			     "handle\n");
		goto disc;
	}

	/* move QP state to RTR and RTS */
	dapl_dbg_log(DAPL_DBG_TYPE_CM, 
		    " rep_recv: RTR_RTS: id %d rqp %x rlid %x rSID %d\n",
		     conn->cm_id,event->param.rep_rcvd.remote_qpn,
		     ntohs(conn->req.primary_path->dlid),conn->service_id);

	if ( dapls_modify_qp_state( conn->ep->qp_handle, 
				    IBV_QPS_RTR, conn ) != DAT_SUCCESS )
		goto disc;

	if ( dapls_modify_qp_state( conn->ep->qp_handle, 
				    IBV_QPS_RTS, conn ) != DAT_SUCCESS)
		goto disc; 


	status = ib_cm_send_rtu(conn->cm_id, NULL, 0);
	if (status) {
		dapl_dbg_log(DAPL_DBG_TYPE_ERR, 
			     " dapli_rep_recv: ib_send_cm_rtu "
			     "failed: %d\n", status);
		goto disc;
	}
	dapl_dbg_log(DAPL_DBG_TYPE_CM, 
     		     " rep_recv: conn %p id %d CONNECTED!\n",
     		     conn, conn->cm_id );

	dapl_evd_connection_callback(conn, IB_CME_CONNECTED,
				     event->private_data, conn->ep);
	return;

disc:
	dapl_evd_connection_callback(conn, IB_CME_LOCAL_FAILURE, NULL,
				     conn->ep);
}

static struct dapl_cm_id * dapli_req_recv(struct dapl_cm_id  *conn,
					  struct ib_cm_event *event)
{
	struct dapl_cm_id 	*new_conn;

	if (conn->sp == NULL) {
		dapl_dbg_log(DAPL_DBG_TYPE_ERR, 
			     " dapli_rep_recv: on invalid listen "
			     "handle\n");
		return NULL;
	}

	/* allocate new cm_id and merge listen parameters */
	new_conn = dapl_os_alloc(sizeof(*new_conn)); 
	if (new_conn) {

		(void)dapl_os_memzero(new_conn, sizeof(*new_conn));
		new_conn->cm_id = event->cm_id; /* provided by uCM */
		event->cm_id->context = new_conn; /* update CM_ID context */
		new_conn->sp = conn->sp;
		new_conn->hca = conn->hca;
		new_conn->service_id = conn->service_id;

		/* save request information in new conn */
		dapl_os_memcpy(&new_conn->req_rcvd, 
				&event->param.req_rcvd,
				sizeof(struct ib_cm_req_event_param));

		new_conn->req_rcvd.primary_path = &new_conn->dapl_path;
		new_conn->req_rcvd.alternate_path = NULL;

		dapl_os_memcpy(new_conn->req_rcvd.primary_path,
				event->param.req_rcvd.primary_path,
				sizeof(struct ib_sa_path_rec));
			 	
		dapl_dbg_log(DAPL_DBG_TYPE_CM, " passive_cb: "
			"REQ on HCA %p SP %p SID %d LID %d new_id %d pd %p\n",
			new_conn->hca, new_conn->sp, 
			conn->service_id, conn->cm_id, new_conn->cm_id, 
			event->private_data );

	}
	return new_conn;
}


static int dapli_cm_active_cb(struct dapl_cm_id *conn,
			       struct ib_cm_event *event)
{
	int destroy;

	dapl_dbg_log(DAPL_DBG_TYPE_CM, 
		     " active_cb: conn %p id %d event %d\n",
		     conn, conn->cm_id, event->event );

	dapl_os_lock(&conn->lock);
	if (conn->destroy) {
		dapl_os_unlock(&conn->lock);
		return 0;
	}
	conn->in_callback = 1;
	dapl_os_unlock(&conn->lock);

	switch (event->event) {
	case IB_CM_REQ_ERROR:
		dapl_evd_connection_callback(conn,
					     IB_CME_DESTINATION_UNREACHABLE,
					     NULL, conn->ep);
		break;
	case IB_CM_REJ_RECEIVED:
		dapl_evd_connection_callback(conn, IB_CME_DESTINATION_REJECT,
					     NULL, conn->ep);
		break;
	case IB_CM_REP_RECEIVED:
		dapli_rep_recv(conn, event);
		break;
	case IB_CM_DREQ_RECEIVED:
		ib_cm_send_drep(conn->cm_id, NULL, 0);
		break;
	case IB_CM_DREQ_ERROR:
	case IB_CM_DREP_RECEIVED:
		/* Wait to exit timewait. */
		break;
	case IB_CM_TIMEWAIT_EXIT:
		dapl_evd_connection_callback(conn, IB_CME_DISCONNECTED,
					     NULL, conn->ep);
		break;
	default:
		dapl_dbg_log(DAPL_DBG_TYPE_ERR,
			     " dapli_cm_active_cb_handler: Unexpected CM "
			     "event %d on ID 0x%p\n", event->event, conn->cm_id);
		break;
	}

	dapl_os_lock(&conn->lock);
	destroy = conn->destroy;
	conn->in_callback = conn->destroy;
	dapl_os_unlock(&conn->lock);
	if (destroy) {
		dapl_dbg_log(DAPL_DBG_TYPE_CM, 
			     " active_cb: DESTROY conn %p id %d \n",
			     conn, conn->cm_id );
		if (conn->ep)
			conn->ep->cm_handle = IB_INVALID_HANDLE;
		
		dapl_os_free(conn, sizeof(*conn));
	}
	return(destroy);
}

static int dapli_cm_passive_cb(struct dapl_cm_id *conn,
			       struct ib_cm_event *event)
{
	int destroy;
	struct dapl_cm_id *new_conn;

	dapl_dbg_log(DAPL_DBG_TYPE_CM, 
		     " passive_cb: conn %p id %d event %d\n",
		     conn, conn->cm_id, event->event );

	dapl_os_lock(&conn->lock);
	if (conn->destroy) {
		dapl_os_unlock(&conn->lock);
		return 0;
	}
	conn->in_callback = 1;
	dapl_os_unlock(&conn->lock);

	switch (event->event) {
	case IB_CM_REQ_RECEIVED:
		/* create new conn object with new conn_id from event */
		new_conn = dapli_req_recv(conn,event);

		if (new_conn)	
			dapls_cr_callback(new_conn, 
					  IB_CME_CONNECTION_REQUEST_PENDING, 
				 	  event->private_data, new_conn->sp);
		break;
	case IB_CM_REP_ERROR:
		dapls_cr_callback(conn, IB_CME_DESTINATION_UNREACHABLE,
				 NULL, conn->sp);
		break;
	case IB_CM_REJ_RECEIVED:
		dapls_cr_callback(conn, IB_CME_DESTINATION_REJECT, NULL,
				 conn->sp);
		break;
	case IB_CM_RTU_RECEIVED:
		/* move QP to RTS state */
		if ( dapls_modify_qp_state(conn->ep->qp_handle, 
					   IBV_QPS_RTS, conn ) != DAT_SUCCESS) {
			dapls_cr_callback(conn, IB_CME_LOCAL_FAILURE, 
					  NULL, conn->sp);
		} else {
			dapl_dbg_log(DAPL_DBG_TYPE_CM, 
		     		     " passive_cb: conn %p id %d CONNECTED!\n",
		     		     conn, conn->cm_id );
			dapls_cr_callback(conn, IB_CME_CONNECTED, 
					  NULL, conn->sp);
		}
		break;
	case IB_CM_DREQ_RECEIVED:
		ib_cm_send_drep(conn->cm_id, NULL, 0);
		break;
	case IB_CM_DREQ_ERROR:
	case IB_CM_DREP_RECEIVED:
		/* Wait to exit timewait. */
		break;
	case IB_CM_TIMEWAIT_EXIT:
		dapls_cr_callback(conn, IB_CME_DISCONNECTED, NULL, conn->sp);
		break;
	default:
		dapl_dbg_log(DAPL_DBG_TYPE_ERR, " dapl_cm_passive_cb_handler: "
			     "Unexpected CM event %d on ID 0x%p\n",
			     event->event, conn->cm_id);
		break;
	}

	dapl_os_lock(&conn->lock);
	destroy = conn->destroy;
	conn->in_callback = conn->destroy;
	dapl_os_unlock(&conn->lock);
	if (destroy) {
		if (conn->ep)
			conn->ep->cm_handle = IB_INVALID_HANDLE;

		dapl_os_free(conn, sizeof(*conn));
	}
	return(destroy);
}


/************************ DAPL provider entry points **********************/

/*
 * dapls_ib_connect
 *
 * Initiate a connection with the passive listener on another node
 *
 * Input:
 *	ep_handle,
 *	remote_ia_address,
 *	remote_conn_qual,
 *	prd_size		size of private data and structure
 *	prd_prt			pointer to private data structure
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *	DAT_INVALID_PARAMETER
 *
 */
DAT_RETURN
dapls_ib_connect (
	IN  DAT_EP_HANDLE		ep_handle,
	IN  DAT_IA_ADDRESS_PTR		r_addr,
	IN  DAT_CONN_QUAL		r_qual,
	IN  DAT_COUNT			p_size,
	IN  void			*p_data )
{
	DAPL_EP		*ep_ptr;
	ib_qp_handle_t	qp_ptr;
	int		status;
	DAT_RETURN	dat_status = DAT_INTERNAL_ERROR;
	ib_cm_handle_t	conn;
	
	/*
	 *  Sanity check
	 */
	if ( NULL == ep_handle ) 
		return DAT_SUCCESS;

	ep_ptr = (DAPL_EP*)ep_handle;
	qp_ptr = ep_ptr->qp_handle;

	dapl_dbg_log(DAPL_DBG_TYPE_CM, " connect: rSID %d, pdata %p, ln %d\n", 
		     r_qual,p_data,p_size);
			
	/* Allocate CM and initialize lock */
	if ((conn = dapl_os_alloc(sizeof(*conn))) == NULL) 
		return DAT_INSUFFICIENT_RESOURCES;
	
	(void)dapl_os_memzero(conn, sizeof(*conn));

	dat_status = dapl_os_lock_init(&conn->lock);
	if (dat_status != DAT_SUCCESS) { 
		dapl_os_free(conn, sizeof(*conn));
		return DAT_INTERNAL_ERROR;
	}

	conn->ep = ep_ptr;
	conn->hca = ep_ptr->header.owner_ia->hca_ptr;
	status = ib_cm_create_id(conn->hca->ib_hca_handle,
				 &conn->cm_id, conn);
	if (status < 0)  {
		dat_status = dapl_convert_errno(errno,"create_cm_id");
		dapl_os_free(conn, sizeof(*conn));
		return dat_status;
	}
	ep_ptr->cm_handle = conn;

	/* Setup QP/CM parameters */
	(void)dapl_os_memzero(&conn->req,sizeof(conn->req));
	conn->service_id = r_qual;
	conn->req.qp_num = ep_ptr->qp_handle->qp_num;
	conn->req.qp_type = IBV_QPT_RC;
	conn->req.starting_psn = ep_ptr->qp_handle->qp_num;
	conn->req.private_data = p_data;
	conn->req.private_data_len = p_size;
	conn->req.peer_to_peer = 0;
	conn->req.responder_resources = IB_TARGET_MAX;
	conn->req.initiator_depth = IB_INITIATOR_DEPTH;
	conn->req.remote_cm_response_timeout = IB_CM_RESPONSE_TIMEOUT;
	conn->req.flow_control = 1;
	conn->req.local_cm_response_timeout = IB_CM_RESPONSE_TIMEOUT;
	conn->req.retry_count = IB_RC_RETRY_COUNT;
	conn->req.rnr_retry_count = IB_RNR_RETRY_COUNT;
	conn->req.max_cm_retries = IB_MAX_CM_RETRIES;
	conn->req.srq = 0;
	
	conn->dapl_comp.fn = &dapli_rt_comp_handler;
	conn->dapl_comp.context = conn;
	conn->retries = 0;
	dapl_os_memcpy(&conn->r_addr, r_addr, sizeof(DAT_SOCK_ADDR6));
	
	status = ib_at_route_by_ip(
		((struct sockaddr_in *)&conn->r_addr)->sin_addr.s_addr, 
		((struct sockaddr_in *)&conn->hca->hca_address)->sin_addr.s_addr,
		0, 0, &conn->dapl_rt, &conn->dapl_comp, &conn->dapl_comp.req_id);

	dapl_dbg_log(DAPL_DBG_TYPE_CM, 
		" connect: at_route requested(ret=%d,id=%d): SRC %x DST %x\n", 
	     status, conn->dapl_comp.req_id,
	     ntohl(((struct sockaddr_in *)&conn->hca->hca_address)->sin_addr.s_addr),
	     ntohl(((struct sockaddr_in *)&conn->r_addr)->sin_addr.s_addr));

	if (status < 0) {
		dat_status = dapl_convert_errno(errno,"ib_at_route_by_ip");
		dapli_destroy_cm_id(conn); 
		return dat_status;
	}

	if (status > 0) 
		dapli_rt_comp_handler(conn->dapl_comp.req_id, conn, status);

	return DAT_SUCCESS;
}

/*
 * dapls_ib_disconnect
 *
 * Disconnect an EP
 *
 * Input:
 *	ep_handle,
 *	disconnect_flags
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *
 */
DAT_RETURN
dapls_ib_disconnect (
	IN	DAPL_EP			*ep_ptr,
	IN	DAT_CLOSE_FLAGS		close_flags )
{
	ib_cm_handle_t	conn = ep_ptr->cm_handle;
	int status;

	dapl_dbg_log (DAPL_DBG_TYPE_CM,
			" disconnect(ep %p, conn %p, id %d flags %x)\n",
			ep_ptr,conn, (conn?conn->cm_id:0),close_flags);

	if (conn == IB_INVALID_HANDLE)
		return DAT_SUCCESS;

	if (close_flags == DAT_CLOSE_ABRUPT_FLAG)
		dapli_destroy_cm_id(conn);
	else {
		status = ib_cm_send_dreq(conn->cm_id, NULL, 0);
		if (status)
			dapl_dbg_log(DAPL_DBG_TYPE_ERR,
				     "dapl_ib_disconnect: ID %p status %d\n", 
				     ep_ptr->cm_handle, status);
	}

	dapl_ep_legacy_post_disconnect(ep_ptr, close_flags)

	return DAT_SUCCESS;
}

/*
 * dapls_ib_disconnect_clean
 *
 * Clean up outstanding connection data. This routine is invoked
 * after the final disconnect callback has occurred. Only on the
 * ACTIVE side of a connection.
 *
 * Input:
 *	ep_ptr		DAPL_EP
 *	active		Indicates active side of connection
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	void
 *
 */
void
dapls_ib_disconnect_clean (
	IN  DAPL_EP			*ep_ptr,
	IN  DAT_BOOLEAN			active,
	IN  const ib_cm_events_t	ib_cm_event )
{

	/*
	 * Clean up outstanding connection state
	 */
	dapls_ib_disconnect(ep_ptr, DAT_CLOSE_ABRUPT_FLAG);
         
	if (ep_ptr->qp_handle != IB_INVALID_HANDLE) 
		dapls_modify_qp_state(ep_ptr->qp_handle, IBV_QPS_ERR, 0 );
}

/*
 * dapl_ib_setup_conn_listener
 *
 * Have the CM set up a connection listener.
 *
 * Input:
 *	ibm_hca_handle		HCA handle
 *	qp_handle			QP handle
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *	DAT_INTERNAL_ERROR
 *	DAT_CONN_QUAL_UNAVAILBLE
 *	DAT_CONN_QUAL_IN_USE
 *
 */
DAT_RETURN
dapls_ib_setup_conn_listener (
	IN  DAPL_IA		*ia_ptr,
	IN  DAT_UINT64		ServiceID,
	IN  DAPL_SP		*sp_ptr )
{
	DAT_RETURN		dat_status = DAT_SUCCESS;
	int			status;
	ib_cm_srvc_handle_t     conn;


	/* Allocate CM and initialize lock */
	if ((conn = dapl_os_alloc(sizeof(*conn))) == NULL) 
		return DAT_INSUFFICIENT_RESOURCES;
	
	(void)dapl_os_memzero(conn, sizeof(*conn));

	dat_status = dapl_os_lock_init(&conn->lock);
	if (dat_status != DAT_SUCCESS)  {
		dapl_os_free(conn, sizeof(*conn));
		return DAT_INTERNAL_ERROR;
	}
	 
	status = ib_cm_create_id(ia_ptr->hca_ptr->ib_hca_handle,
				 &conn->cm_id, conn);
	if (status < 0)  {
		dat_status = dapl_convert_errno(errno,"create_cm_id");
		dapl_os_free(conn, sizeof(*conn));
		return dat_status;
	}

	dapl_dbg_log(DAPL_DBG_TYPE_CM,
		" setup_listener(ia_ptr %p SID %d sp %p conn %p id %d)\n",
		ia_ptr, ServiceID, sp_ptr, conn, conn->cm_id );

	sp_ptr->cm_srvc_handle = conn;
	conn->sp = sp_ptr;
	conn->hca = ia_ptr->hca_ptr;
	conn->service_id = ServiceID;

	dapl_dbg_log(DAPL_DBG_TYPE_EP,
		     " setup_listener(conn=%p cm_id=%d)\n",
		     sp_ptr->cm_srvc_handle,conn->cm_id);

	status = ib_cm_listen(conn->cm_id,
			      cpu_to_be64(ServiceID), 0);
	if (status) {
		if (status == -EBUSY)
			dat_status = DAT_CONN_QUAL_IN_USE;
		else
			dat_status = DAT_INSUFFICIENT_RESOURCES;

        	dapli_destroy_cm_id(conn);
		return dat_status;
	}

	/* success */ 
	return DAT_SUCCESS;
}


/*
 * dapl_ib_remove_conn_listener
 *
 * Have the CM remove a connection listener.
 *
 * Input:
 *	ia_handle		IA handle
 *	ServiceID		IB Channel Service ID
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INVALID_STATE
 *
 */
DAT_RETURN
dapls_ib_remove_conn_listener (
	IN  DAPL_IA		*ia_ptr,
	IN  DAPL_SP		*sp_ptr )
{
	ib_cm_srvc_handle_t	conn = sp_ptr->cm_srvc_handle;

	dapl_dbg_log (DAPL_DBG_TYPE_CM,
			" remove_listener(ia_ptr %p sp_ptr %p cm_ptr %p)\n",
			ia_ptr, sp_ptr, conn );
	
	if (conn != IB_INVALID_HANDLE) { 
		sp_ptr->cm_srvc_handle = NULL;
        	dapli_destroy_cm_id(conn);
	}	

	return DAT_SUCCESS;
}

/*
 * dapls_ib_accept_connection
 *
 * Perform necessary steps to accept a connection
 *
 * Input:
 *	cr_handle
 *	ep_handle
 *	private_data_size
 *	private_data
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *	DAT_INTERNAL_ERROR
 *
 */
DAT_RETURN
dapls_ib_accept_connection (
	IN  DAT_CR_HANDLE	cr_handle,
	IN  DAT_EP_HANDLE	ep_handle,
	IN  DAT_COUNT		p_size,
	IN  const DAT_PVOID	p_data )
{
	DAPL_CR *cr_ptr;
	DAPL_EP *ep_ptr;
	DAPL_IA *ia_ptr;
	DAT_RETURN dat_status;
	int status;
	struct ib_cm_rep_param passive_params;
	struct dapl_cm_id *conn;

	cr_ptr = (DAPL_CR *) cr_handle;
	ep_ptr = (DAPL_EP *) ep_handle;
	ia_ptr = ep_ptr->header.owner_ia;
	conn = cr_ptr->ib_cm_handle;

	dapl_dbg_log (DAPL_DBG_TYPE_CM,
		      " accept(cr %p conn %p, id %d, p_data %p, p_sz=%d)\n",
		      cr_ptr, conn, conn->cm_id, p_data, p_size );

	/* Obtain size of private data structure & contents */
	if (p_size > IB_MAX_REP_PDATA_SIZE) {
		dat_status = DAT_ERROR(DAT_LENGTH_ERROR, DAT_NO_SUBTYPE);
		goto reject;
	}

	if (ep_ptr->qp_state == DAPL_QP_STATE_UNATTACHED) {
		/* 
		 * If we are lazy attaching the QP then we may need to
		 * hook it up here. Typically, we run this code only for
		 * DAT_PSP_PROVIDER_FLAG
		 */
		dat_status = dapls_ib_qp_alloc(ia_ptr, ep_ptr, NULL);
		if (dat_status != DAT_SUCCESS) {
			dapl_dbg_log(DAPL_DBG_TYPE_ERR,
				     " accept: ib_qp_alloc failed: %d\n",
				     dat_status);
			goto reject;
		}
	}

        /* move QP to RTR state */
	dat_status = dapls_modify_qp_state(ep_ptr->qp_handle, 
					   IBV_QPS_RTR, conn); 
	if (dat_status != DAT_SUCCESS ) {
		dapl_dbg_log(DAPL_DBG_TYPE_ERR,
				     " accept: modify_qp_state failed: %d\n",
				     dat_status);
		goto reject;
	}

	cr_ptr->param.local_ep_handle = ep_handle;
	ep_ptr->cm_handle = conn;
	conn->ep = ep_ptr;

	memset(&passive_params, 0, sizeof(passive_params));
	passive_params.private_data = p_data;
	passive_params.private_data_len = p_size;
	passive_params.qp_num = ep_ptr->qp_handle->qp_num;
	passive_params.starting_psn = ep_ptr->qp_handle->qp_num;
	passive_params.responder_resources = IB_TARGET_MAX;
	passive_params.initiator_depth = IB_INITIATOR_DEPTH;
	passive_params.rnr_retry_count = IB_RNR_RETRY_COUNT;

	status = ib_cm_send_rep(conn->cm_id, &passive_params);
	if (status) {
		dapl_dbg_log(DAPL_DBG_TYPE_ERR, " accept: "
			     "ib_send_cm_rep failed: %d\n", status);
		dat_status = DAT_ERROR(DAT_INSUFFICIENT_RESOURCES, 0);
		goto reject;
	}
	return DAT_SUCCESS;

reject:
	ib_cm_send_rej(conn->cm_id, IB_CM_REJ_CONSUMER_DEFINED, NULL, 0,
		       NULL, 0);
	dapli_destroy_cm_id(conn);
	return dat_status;
   
}


/*
 * dapls_ib_reject_connection
 *
 * Reject a connection
 *
 * Input:
 *	cr_handle
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INTERNAL_ERROR
 *
 */
DAT_RETURN
dapls_ib_reject_connection (
	IN  ib_cm_handle_t	cm_handle,
	IN  int			reject_reason )
{
    	int status;

	dapl_dbg_log (DAPL_DBG_TYPE_CM,
		      " reject_connection(cm_handle %p reason %x)\n",
		      cm_handle, reject_reason );

	if (cm_handle == IB_INVALID_HANDLE) {
		dapl_dbg_log(DAPL_DBG_TYPE_ERR,
			     " reject_conn: invalid handle: reason %d\n",
			     reject_reason);
		return DAT_SUCCESS;
	}

	status = ib_cm_send_rej(cm_handle->cm_id, IB_CM_REJ_CONSUMER_DEFINED,
				NULL, 0, NULL, 0);
	if (status) {
		dapl_dbg_log(DAPL_DBG_TYPE_ERR, 
			     " reject_conn: cm_rej failed: %d\n", status);
		return DAT_ERROR(DAT_INSUFFICIENT_RESOURCES, 0);
	}
	dapli_destroy_cm_id(cm_handle);
	return DAT_SUCCESS;
}

/*
 * dapls_ib_cm_remote_addr
 *
 * Obtain the remote IP address given a connection
 *
 * Input:
 *	cr_handle
 *
 * Output:
 *	remote_ia_address: where to place the remote address
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INVALID_HANDLE
 *
 */
DAT_RETURN
dapls_ib_cm_remote_addr (
	IN      DAT_HANDLE	dat_handle,
	OUT	DAT_SOCK_ADDR6	*remote_ia_address )
{
	DAPL_HEADER	*header;
	ib_cm_handle_t	ib_cm_handle;

	dapl_dbg_log (DAPL_DBG_TYPE_EP,
		      "dapls_ib_cm_remote_addr(dat_handle %p, ....)\n",
		      dat_handle );

	header = (DAPL_HEADER *)dat_handle;

	if (header->magic == DAPL_MAGIC_EP) 
		ib_cm_handle = ((DAPL_EP *) dat_handle)->cm_handle;
	else if (header->magic == DAPL_MAGIC_CR) 
		ib_cm_handle = ((DAPL_CR *) dat_handle)->ib_cm_handle;
	else 
		return DAT_INVALID_HANDLE;

	dapl_os_memcpy(	remote_ia_address, 
			&ib_cm_handle->r_addr, 
			sizeof(DAT_SOCK_ADDR6) );

	return DAT_SUCCESS;
}

/*
 * dapls_ib_private_data_size
 *
 * Return the size of private data given a connection op type
 *
 * Input:
 *	prd_ptr		private data pointer
 *	conn_op		connection operation type
 *
 * If prd_ptr is NULL, this is a query for the max size supported by
 * the provider, otherwise it is the actual size of the private data
 * contained in prd_ptr.
 *
 *
 * Output:
 *	None
 *
 * Returns:
 * 	length of private data
 *
 */
int dapls_ib_private_data_size (
	IN      DAPL_PRIVATE	*prd_ptr,
	IN	DAPL_PDATA_OP	conn_op)
{
	int  size;

	switch (conn_op)
	{
		case DAPL_PDATA_CONN_REQ:
		{
			size = IB_MAX_REQ_PDATA_SIZE;
			break;
		}
		case DAPL_PDATA_CONN_REP:
		{
			size = IB_MAX_REP_PDATA_SIZE;
			break;
		}
		case DAPL_PDATA_CONN_REJ:
		{
			size = IB_MAX_REJ_PDATA_SIZE;
			break;
		}
		case DAPL_PDATA_CONN_DREQ:
		{
			size = IB_MAX_DREQ_PDATA_SIZE;
			break;
		}
		case DAPL_PDATA_CONN_DREP:
		{
			size = IB_MAX_DREP_PDATA_SIZE;
			break;
		}
		default:
		{
			size = 0;
		}

	} /* end case */

	return size;
}

/*
 * Map all socket CM event codes to the DAT equivelent.
 */
#define DAPL_IB_EVENT_CNT	12

static struct ib_cm_event_map
{
	const ib_cm_events_t	ib_cm_event;
	DAT_EVENT_NUMBER	dat_event_num;
	} ib_cm_event_map[DAPL_IB_EVENT_CNT] = {
	/* 00 */  { IB_CME_CONNECTED,	
					DAT_CONNECTION_EVENT_ESTABLISHED}, 
	/* 01 */  { IB_CME_DISCONNECTED,	
					DAT_CONNECTION_EVENT_DISCONNECTED},
	/* 02 */  { IB_CME_DISCONNECTED_ON_LINK_DOWN, 
					DAT_CONNECTION_EVENT_DISCONNECTED},
	/* 03 */  { IB_CME_CONNECTION_REQUEST_PENDING,	
					DAT_CONNECTION_REQUEST_EVENT},
	/* 04 */  { IB_CME_CONNECTION_REQUEST_PENDING_PRIVATE_DATA,
					DAT_CONNECTION_REQUEST_EVENT},
	/* 05 */  { IB_CME_CONNECTION_REQUEST_ACKED,
					DAT_CONNECTION_REQUEST_EVENT},
	/* 06 */  { IB_CME_DESTINATION_REJECT,
					DAT_CONNECTION_EVENT_NON_PEER_REJECTED},
	/* 07 */  { IB_CME_DESTINATION_REJECT_PRIVATE_DATA,		
					DAT_CONNECTION_EVENT_PEER_REJECTED},
	/* 08 */  { IB_CME_DESTINATION_UNREACHABLE,	
					DAT_CONNECTION_EVENT_UNREACHABLE},
	/* 09 */  { IB_CME_TOO_MANY_CONNECTION_REQUESTS,
					DAT_CONNECTION_EVENT_NON_PEER_REJECTED},
	/* 10 */  { IB_CME_LOCAL_FAILURE,
					DAT_CONNECTION_EVENT_BROKEN},
	/* 11 */  { IB_CME_BROKEN,
					DAT_CONNECTION_EVENT_BROKEN}
};
 
/*
 * dapls_ib_get_cm_event
 *
 * Return a DAT connection event given a provider CM event.
 *
 * Input:
 *	dat_event_num	DAT event we need an equivelent CM event for
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	ib_cm_event of translated DAPL value
 */
DAT_EVENT_NUMBER
dapls_ib_get_dat_event (
	IN    const ib_cm_events_t	ib_cm_event,
	IN    DAT_BOOLEAN		active)
{
	DAT_EVENT_NUMBER	dat_event_num;
	int			i;
	
	active = active;

	if (ib_cm_event > IB_CME_BROKEN)
		return (DAT_EVENT_NUMBER) 0;

	dat_event_num = 0;
	for (i = 0; i < DAPL_IB_EVENT_CNT; i++) {
		if (ib_cm_event == ib_cm_event_map[i].ib_cm_event) {
			dat_event_num = ib_cm_event_map[i].dat_event_num;
			break;
		}
	}
	dapl_dbg_log (DAPL_DBG_TYPE_CALLBACK,
		"dapls_ib_get_dat_event: event(%s) ib=0x%x dat=0x%x\n",
		active ? "active" : "passive", ib_cm_event, dat_event_num);

	return dat_event_num;
}


/*
 * dapls_ib_get_dat_event
 *
 * Return a DAT connection event given a provider CM event.
 * 
 * Input:
 *	ib_cm_event	event provided to the dapl callback routine
 *	active		switch indicating active or passive connection
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_EVENT_NUMBER of translated provider value
 */
ib_cm_events_t
dapls_ib_get_cm_event (
	IN    DAT_EVENT_NUMBER		dat_event_num)
{
    ib_cm_events_t	ib_cm_event;
    int			i;

    ib_cm_event = 0;
    for (i = 0; i < DAPL_IB_EVENT_CNT; i++) {
	if ( dat_event_num == ib_cm_event_map[i].dat_event_num ) {
		ib_cm_event = ib_cm_event_map[i].ib_cm_event;
		break;
	}
    }
    return ib_cm_event;
}


void dapli_cm_event_cb(struct _ib_hca_transport *hca)
{
	struct ib_cm_event *event;
		
	dapl_dbg_log(DAPL_DBG_TYPE_UTIL, " dapli_cm_event()\n");

	/* process one CM event, fairness */
	if(!ib_cm_get_event(hca->ib_cm,&event)) {
		struct dapl_cm_id	*conn;
		int			ret;
		dapl_dbg_log(DAPL_DBG_TYPE_CM,
			     " dapli_cm_event: EVENT=%p ID=%p CTX=%p\n",
			     event->event, event->cm_id, 
			     event->cm_id->context);
		
		/* set proper conn from cm_id context*/
		conn = (struct dapl_cm_id*)event->cm_id->context;

		if (conn->sp) 
			ret = dapli_cm_passive_cb(conn,event);
		else 
			ret = dapli_cm_active_cb(conn,event);
		
		ib_cm_ack_event(event);

		if (ret) 
			ib_cm_destroy_id(conn->cm_id);
	}
}

void dapli_at_event_cb()
{
	dapl_dbg_log(DAPL_DBG_TYPE_UTIL, " dapli_at_event_cb()\n");

	/* process one AT event, fairness */
	ib_at_callback_get_timed(0);
}


/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 8
 * End:
 */
