/*
 * This Software is licensed under one of the following licenses:
 *
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/cpl.php.
 *
 * 2) under the terms of the "The BSD License" a copy of which is
 *    available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 *
 * 3) under the terms of the "GNU General Public License (GPL) Version 2" a
 *    copy of which is available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/gpl-license.php.
 *
 * Licensee has the right to choose one of the above licenses.
 *
 * Redistributions of source code must retain the above copyright
 * notice and one of the license notices.
 *
 * Redistributions in binary form must reproduce both the above copyright
 * notice, one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/***************************************************************************
 *
 *   Module:		 uDAPL
 *
 *   Filename:		 dapl_ib_util.c
 *
 *   Author:		 Arlin Davis
 *
 *   Created:		 3/10/2005
 *
 *   Description: 
 *
 *   The uDAPL openib provider - init, open, close, utilities, work thread
 *
 ****************************************************************************
 *		   Source Control System Information
 *
 *    $Id: $
 *
 *	Copyright (c) 2005 Intel Corporation.  All rights reserved.
 *
 **************************************************************************/
#ifdef RCSID
static const char rcsid[] = "$Id:  $";
#endif

#include "dapl.h"
#include "dapl_adapter_util.h"
#include "dapl_ib_util.h"

#include <stdlib.h>
#include <netinet/tcp.h>
#include <sys/poll.h>
#include <fcntl.h>

int			g_dapl_loopback_connection = 0;
int			g_ib_destroy = 0;
int			g_ib_pipe[2];
DAPL_OS_THREAD		g_ib_thread;
DAPL_OS_LOCK		g_hca_lock;
struct dapl_llist_entry	*g_hca_list;	

/* just get IP address, IPv4 only for now  */
int dapli_get_hca_addr( struct dapl_hca *hca_ptr )
{
	struct sockaddr_in	*ipv4_addr;
	struct ib_at_completion	at_comp;	
	struct dapl_at_record	at_rec;
	int			status;
	DAT_RETURN		dat_status;
	
	ipv4_addr = (struct sockaddr_in*)&hca_ptr->hca_address;
	ipv4_addr->sin_family = AF_INET;
	ipv4_addr->sin_addr.s_addr = 0;
	
	at_comp.fn = dapli_ip_comp_handler;
	at_comp.context = &at_rec; 
	at_rec.addr = &hca_ptr->hca_address;
	at_rec.wait_object = &hca_ptr->ib_trans.wait_object;
	at_rec.hca_ptr = hca_ptr;
	at_rec.retries = 0;

	/*  call with async_comp until the sync version works */
	status = ib_at_ips_by_gid(&hca_ptr->ib_trans.gid, 
				  &ipv4_addr->sin_addr.s_addr, 1, 
				  &at_comp, &at_rec.req_id);
	
	if (status < 0) {
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
			      " ips_by_gid: ERR ips_by_gid %d %s \n",
				status, strerror(errno));
		return 1;
	}
 
	dapl_dbg_log (DAPL_DBG_TYPE_UTIL, 
		      " ips_by_gid: RET %d at_rec %p -> id %lld\n",
			status, &at_rec, at_rec.req_id );

        if (status > 0) { 
                dapli_ip_comp_handler(at_rec.req_id, (void*)&at_rec, status);
	} else {
		/* limit the resolution and cancel times */
		dat_status = dapl_os_wait_object_wait(
					&hca_ptr->ib_trans.wait_object,
					500000);
		if (dat_status != DAT_SUCCESS) {
			dapl_dbg_log(
				DAPL_DBG_TYPE_UTIL, 
				" ips_by_gid: REQ TIMEOUT, cancel %lld\n",
				at_rec.req_id);
			
			/* 
			 * FIX: indeterministic
			 * AT may or may not provide -EINTR event 
			 */
			ib_at_cancel(at_rec.req_id); 
			
			if (dapl_os_wait_object_wait(
					&hca_ptr->ib_trans.wait_object,
					500000) != DAT_SUCCESS)
			{
				dapl_dbg_log(DAPL_DBG_TYPE_ERR, 
					" ips_by_gid: cancel %lld failed\n",
					at_rec.req_id);
			}
		}
	}

	if (!ipv4_addr->sin_addr.s_addr) 
		return 1;
		
	return 0;
}

/*
 * dapls_ib_init, dapls_ib_release
 *
 * Initialize Verb related items for device open
 *
 * Input:
 * 	none
 *
 * Output:
 *	none
 *
 * Returns:
 * 	0 success, -1 error
 *
 */
int32_t dapls_ib_init (void)
{	
	long	opts;
	dapl_dbg_log (DAPL_DBG_TYPE_UTIL, " dapl_ib_init: \n" );

	/* initialize hca_list lock */
	dapl_os_lock_init(&g_hca_lock);
	
	/* initialize hca list for CQ events */
	dapl_llist_init_head(&g_hca_list);

	/* create pipe for waking up work thread */
	if (pipe(g_ib_pipe))
		return 1;

	/* set AT fd to non-blocking */ 
	opts = fcntl(ib_at_get_fd(), F_GETFL);
	if (fcntl(ib_at_get_fd(), F_SETFL, opts | O_NONBLOCK) < 0)
		return 1;

	if (dapli_ib_thread_init()) 
		return 1;

	return 0;
}

int32_t dapls_ib_release (void)
{
	dapl_dbg_log (DAPL_DBG_TYPE_UTIL, " dapl_ib_release: \n" );
	dapli_ib_thread_destroy();
	return 0;
}

/*
 * dapls_ib_open_hca
 *
 * Open HCA
 *
 * Input:
 *      *hca_name         pointer to provider device name
 *      *ib_hca_handle_p  pointer to provide HCA handle
 *
 * Output:
 *      none
 *
 * Return:
 *      DAT_SUCCESS
 *      dapl_convert_errno
 *
 */
DAT_RETURN dapls_ib_open_hca (
        IN   IB_HCA_NAME	hca_name,
        IN   DAPL_HCA		*hca_ptr)
{
	struct ibv_device **dev_list;
	long		opts;
	int		i;
	dapl_dbg_log (DAPL_DBG_TYPE_UTIL, 
		      " open_hca: %s - %p\n", hca_name, hca_ptr );

	/* Get list of all IB devices, find match, open */
	dev_list = ibv_get_device_list(NULL);
	if (!dev_list) {
		dapl_dbg_log (DAPL_DBG_TYPE_ERR,
			      " open_hca: ibv_get_device_list() failed\n",
			      hca_name);
		return DAT_INTERNAL_ERROR;
	}

	for (i = 0; dev_list[i]; ++i) {
		hca_ptr->ib_trans.ib_dev = dev_list[i];
		if (!strcmp(ibv_get_device_name(hca_ptr->ib_trans.ib_dev),
						hca_name))
			goto found;
	}

	dapl_dbg_log (DAPL_DBG_TYPE_ERR,
		      " open_hca: IB device %s not found\n",
		      hca_name);
	goto err;
	
found:
	dapl_dbg_log (
	    DAPL_DBG_TYPE_UTIL," open_hca: Found dev %s %016llx\n", 
	    ibv_get_device_name(hca_ptr->ib_trans.ib_dev),
	    (unsigned long long)
		bswap_64(ibv_get_device_guid(hca_ptr->ib_trans.ib_dev)));

	hca_ptr->ib_hca_handle = ibv_open_device(hca_ptr->ib_trans.ib_dev);
	if (!hca_ptr->ib_hca_handle) {
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
			      " open_hca: IB dev open failed for %s\n", 
			      ibv_get_device_name(hca_ptr->ib_trans.ib_dev));
		goto err;
	}
	hca_ptr->ib_trans.ib_ctx = hca_ptr->ib_hca_handle;

	/* set inline max with env or default, get local lid and gid 0 */
	hca_ptr->ib_trans.max_inline_send = 
		dapl_os_get_env_val("DAPL_MAX_INLINE", INLINE_SEND_DEFAULT);

	/* GID with port_num provided, index 0 for now */
	if (ibv_query_gid(hca_ptr->ib_hca_handle, 
			  hca_ptr->port_num, 0, &hca_ptr->ib_trans.gid)) {
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
			      " open_hca: IB get GID failed for %s\n", 
			      ibv_get_device_name(hca_ptr->ib_trans.ib_dev));
		goto bail;
	}
			
	dapl_dbg_log(DAPL_DBG_TYPE_UTIL,
		" open_hca: GID subnet %016llx id %016llx\n",
		(unsigned long long)
			bswap_64(hca_ptr->ib_trans.gid.global.subnet_prefix),
		(unsigned long long)
			bswap_64(hca_ptr->ib_trans.gid.global.interface_id));

	/* get the IP address of the device using GID */
	if (dapli_get_hca_addr(hca_ptr)) {
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
			      " open_hca: ERR ib_at_ips_by_gid for %s\n", 
			      ibv_get_device_name(hca_ptr->ib_trans.ib_dev));
		goto bail;
	}

	/* initialize hca wait object for uAT event */
	dapl_os_wait_object_init(&hca_ptr->ib_trans.wait_object);

	/* set event FD's to non-blocking */
	opts = fcntl(hca_ptr->ib_hca_handle->async_fd, F_GETFL); /* uASYNC */
	if (opts < 0 || fcntl(hca_ptr->ib_hca_handle->async_fd, 
			      F_SETFL, opts | O_NONBLOCK) < 0) {
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
			      " open_hca: ERR with async FD\n" );
		goto bail;
	}

	/* EVD events without direct CQ channels, non-blocking */
	hca_ptr->ib_trans.ib_cq = 
		ibv_create_comp_channel(hca_ptr->ib_hca_handle);
	opts = fcntl(hca_ptr->ib_trans.ib_cq->fd, F_GETFL); /* uCQ */
	if (opts < 0 || fcntl(hca_ptr->ib_trans.ib_cq->fd, 
			      F_SETFL, opts | O_NONBLOCK) < 0) {
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
			      " open_hca: ERR with CQ FD\n" );
		goto bail;
	}

	/* Get CM device handle for events, and set to non-blocking */	
	hca_ptr->ib_trans.ib_cm = ib_cm_get_device(hca_ptr->ib_hca_handle);
	opts = fcntl(hca_ptr->ib_trans.ib_cm->fd, F_GETFL); /* uCM */
	if (opts < 0 || fcntl(hca_ptr->ib_trans.ib_cm->fd, 
			      F_SETFL, opts | O_NONBLOCK) < 0) {
		dapl_dbg_log (DAPL_DBG_TYPE_ERR, 
			      " open_hca: ERR with CM FD\n" );
		goto bail;
	}
	
	/* 
	 * Put new hca_transport on list for async and CQ event processing 
	 * Wakeup work thread to add to polling list
	 */
	dapl_llist_init_entry((DAPL_LLIST_ENTRY*)&hca_ptr->ib_trans.entry);
	dapl_os_lock( &g_hca_lock );
	dapl_llist_add_tail(&g_hca_list, 
			    (DAPL_LLIST_ENTRY*)&hca_ptr->ib_trans.entry, 
			    &hca_ptr->ib_trans.entry);
	write(g_ib_pipe[1], "w", sizeof "w");
	dapl_os_unlock(&g_hca_lock);
	
  	dapl_dbg_log (
		DAPL_DBG_TYPE_UTIL, 
		" open_hca: %s, port %d, %s  %d.%d.%d.%d INLINE_MAX=%d\n",
		ibv_get_device_name(hca_ptr->ib_trans.ib_dev), 
		hca_ptr->port_num,
		((struct sockaddr_in *)
			&hca_ptr->hca_address)->sin_family == AF_INET ?  
			"AF_INET":"AF_INET6",
		((struct sockaddr_in *)
			&hca_ptr->hca_address)->sin_addr.s_addr >> 0 & 0xff,
		((struct sockaddr_in *)
			&hca_ptr->hca_address)->sin_addr.s_addr >> 8 & 0xff,
		((struct sockaddr_in *)
			&hca_ptr->hca_address)->sin_addr.s_addr >> 16 & 0xff,
		((struct sockaddr_in *)
			&hca_ptr->hca_address)->sin_addr.s_addr >> 24 & 0xff,
		hca_ptr->ib_trans.max_inline_send );

	hca_ptr->ib_trans.d_hca = hca_ptr;
	ibv_free_device_list(dev_list);
	return DAT_SUCCESS;

bail:
	ibv_close_device(hca_ptr->ib_hca_handle); 
	hca_ptr->ib_hca_handle = IB_INVALID_HANDLE;
err:
	ibv_free_device_list(dev_list);
	return DAT_INTERNAL_ERROR;
}


/*
 * dapls_ib_close_hca
 *
 * Open HCA
 *
 * Input:
 *      DAPL_HCA   provide CA handle
 *
 * Output:
 *      none
 *
 * Return:
 *      DAT_SUCCESS
 *	dapl_convert_errno 
 *
 */
DAT_RETURN dapls_ib_close_hca (	IN   DAPL_HCA	*hca_ptr )
{
	dapl_dbg_log (DAPL_DBG_TYPE_UTIL," close_hca: %p->%p\n",
			hca_ptr,hca_ptr->ib_hca_handle);

	if (hca_ptr->ib_hca_handle != IB_INVALID_HANDLE) {
		if (ibv_close_device(hca_ptr->ib_hca_handle)) 
			return(dapl_convert_errno(errno,"ib_close_device"));
		hca_ptr->ib_hca_handle = IB_INVALID_HANDLE;
	}

	/* 
	 * Remove hca from async and CQ event processing list
	 * Wakeup work thread to remove from polling list
	 */
	hca_ptr->ib_trans.destroy = 1;
	write(g_ib_pipe[1], "w", sizeof "w");

	/* wait for thread to remove HCA references */
	while (hca_ptr->ib_trans.destroy != 2) {
		struct timespec	sleep, remain;
		sleep.tv_sec = 0;
		sleep.tv_nsec = 10000000; /* 10 ms */
		dapl_dbg_log(DAPL_DBG_TYPE_UTIL, 
			     " ib_thread_destroy: wait on hca %p destroy\n");
		nanosleep (&sleep, &remain);
	}
	return (DAT_SUCCESS);
}
  
/*
 * dapls_ib_query_hca
 *
 * Query the hca attribute
 *
 * Input:
 *	hca_handl		hca handle	
 *	ia_attr			attribute of the ia
 *	ep_attr			attribute of the ep
 *	ip_addr			ip address of DET NIC
 *
 * Output:
 * 	none
 *
 * Returns:
 * 	DAT_SUCCESS
 *	DAT_INVALID_HANDLE
 */

DAT_RETURN dapls_ib_query_hca (
	IN  DAPL_HCA                       *hca_ptr,
	OUT DAT_IA_ATTR                    *ia_attr,
	OUT DAT_EP_ATTR                    *ep_attr,
	OUT DAT_SOCK_ADDR6                 *ip_addr)
{
	struct ibv_device_attr	dev_attr;
	struct ibv_port_attr	port_attr;

	if (hca_ptr->ib_hca_handle == NULL) {
		dapl_dbg_log (DAPL_DBG_TYPE_ERR," query_hca: BAD handle\n");
		return (DAT_INVALID_HANDLE);
	}

	/* local IP address of device, set during ia_open */
	if (ip_addr != NULL)
		memcpy(ip_addr, &hca_ptr->hca_address, sizeof(DAT_SOCK_ADDR6));
	
	if (ia_attr == NULL && ep_attr == NULL) 
		return DAT_SUCCESS;

	/* query verbs for this device and port attributes */	
	if (ibv_query_device(hca_ptr->ib_hca_handle, &dev_attr) ||
			     ibv_query_port(hca_ptr->ib_hca_handle, 
					    hca_ptr->port_num, &port_attr))
		return(dapl_convert_errno(errno,"ib_query_hca"));

	if (ia_attr != NULL) {
		(void) dapl_os_memzero(ia_attr, sizeof(*ia_attr));
		ia_attr->adapter_name[DAT_NAME_MAX_LENGTH - 1] = '\0';
		ia_attr->vendor_name[DAT_NAME_MAX_LENGTH - 1] = '\0';
		ia_attr->ia_address_ptr = 
			(DAT_IA_ADDRESS_PTR)&hca_ptr->hca_address;

		dapl_dbg_log (DAPL_DBG_TYPE_UTIL, 
			" query_hca: %s %s  %d.%d.%d.%d\n", 
			ibv_get_device_name(hca_ptr->ib_trans.ib_dev),
			((struct sockaddr_in *)
			ia_attr->ia_address_ptr)->sin_family == AF_INET ? 
			"AF_INET":"AF_INET6",
			((struct sockaddr_in *)
			ia_attr->ia_address_ptr)->sin_addr.s_addr >> 0 & 0xff,
			((struct sockaddr_in *)
			ia_attr->ia_address_ptr)->sin_addr.s_addr >> 8 & 0xff,
			((struct sockaddr_in *)
			ia_attr->ia_address_ptr)->sin_addr.s_addr >> 16 & 0xff,
			((struct sockaddr_in *)
			ia_attr->ia_address_ptr)->sin_addr.s_addr >> 24 & 0xff);
		
		ia_attr->hardware_version_major   = dev_attr.hw_ver;
		/* ia_attr->hardware_version_minor   = dev_attr.fw_ver; */
		ia_attr->max_eps                  = dev_attr.max_qp;
		ia_attr->max_dto_per_ep           = dev_attr.max_qp_wr;
		ia_attr->max_rdma_read_in         = dev_attr.max_qp_rd_atom;
		ia_attr->max_rdma_read_out        = dev_attr.max_qp_rd_atom;
		ia_attr->max_rdma_read_per_ep_in  = dev_attr.max_qp_rd_atom;
		ia_attr->max_rdma_read_per_ep_out = dev_attr.max_qp_rd_atom;
		ia_attr->max_rdma_read_per_ep_in_guaranteed  = DAT_TRUE;
		ia_attr->max_rdma_read_per_ep_out_guaranteed = DAT_TRUE;
		ia_attr->max_evds                 = dev_attr.max_cq;
		ia_attr->max_evd_qlen             = dev_attr.max_cqe;
		ia_attr->max_iov_segments_per_dto = dev_attr.max_sge;
		ia_attr->max_lmrs                 = dev_attr.max_mr;
		ia_attr->max_lmr_block_size       = dev_attr.max_mr_size;
		ia_attr->max_rmrs                 = dev_attr.max_mw;
		ia_attr->max_lmr_virtual_address  = dev_attr.max_mr_size;
		ia_attr->max_rmr_target_address   = dev_attr.max_mr_size;
		ia_attr->max_pzs                  = dev_attr.max_pd;
		ia_attr->max_mtu_size             = port_attr.max_msg_sz;
		ia_attr->max_rdma_size            = port_attr.max_msg_sz;
		ia_attr->num_transport_attr       = 0;
		ia_attr->transport_attr           = NULL;
		ia_attr->num_vendor_attr          = 0;
		ia_attr->vendor_attr              = NULL;

		dapl_dbg_log (DAPL_DBG_TYPE_UTIL, 
			" query_hca: (%x.%x) ep %d ep_q %d evd %d evd_q %d\n", 
			ia_attr->hardware_version_major,
			ia_attr->hardware_version_minor,
			ia_attr->max_eps, ia_attr->max_dto_per_ep,
			ia_attr->max_evds, ia_attr->max_evd_qlen );
		dapl_dbg_log (DAPL_DBG_TYPE_UTIL, 
			" query_hca: msg %llu rdma %llu iov %d lmr %d rmr %d\n", 
			ia_attr->max_mtu_size, ia_attr->max_rdma_size,
			ia_attr->max_iov_segments_per_dto, ia_attr->max_lmrs, 
			ia_attr->max_rmrs );
	}
	
	if (ep_attr != NULL) {
		(void) dapl_os_memzero(ep_attr, sizeof(*ep_attr));
		ep_attr->max_mtu_size     = port_attr.max_msg_sz;
		ep_attr->max_rdma_size    = port_attr.max_msg_sz;
		ep_attr->max_recv_dtos    = dev_attr.max_qp_wr;
		ep_attr->max_request_dtos = dev_attr.max_qp_wr;
		ep_attr->max_recv_iov     = dev_attr.max_sge;
		ep_attr->max_request_iov  = dev_attr.max_sge;
		ep_attr->max_rdma_read_in = dev_attr.max_qp_rd_atom;
		ep_attr->max_rdma_read_out= dev_attr.max_qp_rd_atom;
		dapl_dbg_log (DAPL_DBG_TYPE_UTIL, 
			" query_hca: MAX msg %llu dto %d iov %d rdma i%d,o%d\n", 
			ep_attr->max_mtu_size,
			ep_attr->max_recv_dtos, ep_attr->max_recv_iov,
			ep_attr->max_rdma_read_in, ep_attr->max_rdma_read_out);
	}

	return DAT_SUCCESS;
}

/*
 * dapls_ib_setup_async_callback
 *
 * Set up an asynchronous callbacks of various kinds
 *
 * Input:
 *	ia_handle		IA handle
 *	handler_type		type of handler to set up
 *	callback_handle 	handle param for completion callbacks
 *	callback		callback routine pointer
 *	context 		argument for callback routine
 *
 * Output:
 *	none
 *
 * Returns:
 *	DAT_SUCCESS
 *	DAT_INSUFFICIENT_RESOURCES
 *	DAT_INVALID_PARAMETER
 *
 */
DAT_RETURN dapls_ib_setup_async_callback (
	IN  DAPL_IA			*ia_ptr,
	IN  DAPL_ASYNC_HANDLER_TYPE	handler_type,
	IN  DAPL_EVD			*evd_ptr,
	IN  ib_async_handler_t		callback,
	IN  void			*context )

{
	ib_hca_transport_t	*hca_ptr;

	dapl_dbg_log (DAPL_DBG_TYPE_UTIL,
			" setup_async_cb: ia %p type %d hdl %p cb %p ctx %p\n",
			ia_ptr, handler_type, evd_ptr, callback, context);

	hca_ptr = &ia_ptr->hca_ptr->ib_trans;
	switch(handler_type)
	{
		case DAPL_ASYNC_UNAFILIATED:
			hca_ptr->async_unafiliated = 
				(ib_async_handler_t)callback;
			hca_ptr->async_un_ctx = context;
			break;
		case DAPL_ASYNC_CQ_ERROR:
			hca_ptr->async_cq_error = 
				(ib_async_cq_handler_t)callback;
			break;
		case DAPL_ASYNC_CQ_COMPLETION:
			hca_ptr->async_cq = 
				(ib_async_dto_handler_t)callback;
			break;
		case DAPL_ASYNC_QP_ERROR:
			hca_ptr->async_qp_error = 
				(ib_async_qp_handler_t)callback;
			break;
		default:
			break;
	}
	return DAT_SUCCESS;
}

int dapli_ib_thread_init(void)
{
	DAT_RETURN dat_status;

	dapl_dbg_log(DAPL_DBG_TYPE_UTIL,
		     " ib_thread_init(%d)\n", getpid());

	/* create thread to process inbound connect request */
	dat_status = dapl_os_thread_create(dapli_thread, NULL, &g_ib_thread);
	if (dat_status != DAT_SUCCESS)
	{
		dapl_dbg_log(DAPL_DBG_TYPE_ERR, 
			     " ib_thread_init: failed to create thread\n");
		return 1;
	}
	return 0;
}

void dapli_ib_thread_destroy(void)
{
	int retries = 10;
	dapl_dbg_log(DAPL_DBG_TYPE_UTIL,
		     " ib_thread_destroy(%d)\n", getpid());
	/* 
	 * wait for async thread to terminate. 
	 * pthread_join would be the correct method
	 * but some applications have some issues
	 */
	 
	/* destroy ib_thread, wait for termination, if not already */
	dapl_os_lock( &g_hca_lock );
	g_ib_destroy = 1;
	write(g_ib_pipe[1], "w", sizeof "w");
		
	while ((g_ib_destroy != 2) && (retries--)) {
		struct timespec	sleep, remain;
		sleep.tv_sec = 0;
		sleep.tv_nsec = 20000000; /* 20 ms */
		dapl_dbg_log(DAPL_DBG_TYPE_UTIL, 
			" ib_thread_destroy: waiting for ib_thread\n");
		write(g_ib_pipe[1], "w", sizeof "w");
		dapl_os_unlock( &g_hca_lock );
		nanosleep(&sleep, &remain);
		dapl_os_lock( &g_hca_lock );
	}
	dapl_os_unlock( &g_hca_lock );
	
	dapl_dbg_log(DAPL_DBG_TYPE_UTIL,
		     " ib_thread_destroy(%d) exit\n",getpid());
}

void dapli_async_event_cb(struct _ib_hca_transport *hca)
{
	struct ibv_async_event	event;
	struct pollfd	async_fd = {
		.fd      = hca->ib_ctx->async_fd,
		.events  = POLLIN,
		.revents = 0
	};
	
	dapl_dbg_log(DAPL_DBG_TYPE_UTIL, " dapli_async_event_cb(%p)\n",hca);

	if (hca->destroy)
		return;

	if ((poll(&async_fd, 1, 0)==1) &&
		(!ibv_get_async_event(hca->ib_ctx, &event))) {

		switch (event.event_type) {

			case	IBV_EVENT_CQ_ERR:
			{
				struct dapl_ep *evd_ptr = 
					event.element.cq->cq_context;

				dapl_dbg_log(
					DAPL_DBG_TYPE_WARN,
					" dapli_async_event CQ (%p) ERR %d\n",
					evd_ptr, event.event_type);				
				
				/* report up if async callback still setup */
				if (hca->async_cq_error)
					hca->async_cq_error(hca->ib_ctx,
							    event.element.cq,	
							    &event,
							    (void*)evd_ptr);
				break;
			}
			case	IBV_EVENT_COMM_EST:
			{
				/* Received msgs on connected QP before RTU */
				struct dapl_ep *ep_ptr = 
					event.element.qp->qp_context;
				
				/* TODO: cannot process COMM_EST until ibv  
				 * guarantees valid QP context for events. 
				 * Race conditions exist with QP destroy call. 
				 * For now, assume the RTU will arrive.
				 */
				dapl_dbg_log(
					DAPL_DBG_TYPE_UTIL,
					" dapli_async_event COMM_EST(qp=%p)\n",
					event.element.qp);	

				if (!DAPL_BAD_HANDLE(ep_ptr, DAPL_MAGIC_EP) &&
				    ep_ptr->cm_handle != IB_INVALID_HANDLE)
					ib_cm_establish(
						ep_ptr->cm_handle->cm_id);
			
				break;
			}
			case	IBV_EVENT_QP_FATAL:
			case	IBV_EVENT_QP_REQ_ERR:
			case	IBV_EVENT_QP_ACCESS_ERR:
			case	IBV_EVENT_QP_LAST_WQE_REACHED:
			case	IBV_EVENT_SRQ_ERR:
			case	IBV_EVENT_SRQ_LIMIT_REACHED:
			case	IBV_EVENT_SQ_DRAINED:
			{
				struct dapl_ep *ep_ptr = 
					event.element.qp->qp_context;

				dapl_dbg_log(
					DAPL_DBG_TYPE_WARN,
					" dapli_async_event QP (%p) ERR %d\n",
					ep_ptr, event.event_type);	
				
				/* report up if async callback still setup */
				if (hca->async_qp_error)
					hca->async_qp_error(hca->ib_ctx,
							    event.element.qp,
							    &event,
							    (void*)ep_ptr);
				break;
			}
			case	IBV_EVENT_PATH_MIG:
			case	IBV_EVENT_PATH_MIG_ERR:
			case	IBV_EVENT_DEVICE_FATAL:
			case	IBV_EVENT_PORT_ACTIVE:
			case	IBV_EVENT_PORT_ERR:
			case	IBV_EVENT_LID_CHANGE:
			case	IBV_EVENT_PKEY_CHANGE:
			case	IBV_EVENT_SM_CHANGE:
			{
				dapl_dbg_log(DAPL_DBG_TYPE_WARN,
					     " dapli_async_event DEV ERR %d\n",
					     event.event_type);	

				/* report up if async callback still setup */
				if (hca->async_unafiliated)
					hca->async_unafiliated( 
							hca->ib_ctx,
							&event,
							hca->async_un_ctx);
				break;
			}
			default:
			{
				dapl_dbg_log (DAPL_DBG_TYPE_WARN,
					      "--> DsEventCb: UNKNOWN\n");
				break;
			}
		}
		ibv_ack_async_event(&event);
	}
}

/* work thread for uAT, uCM, CQ, and async events */
void dapli_thread(void *arg) 
{
	struct pollfd		 ufds[__FD_SETSIZE];
	struct _ib_hca_transport *uhca[__FD_SETSIZE]={NULL};
	struct _ib_hca_transport *hca;
	int			 ret,idx,fds;
	char			 rbuf[2];
	
	dapl_dbg_log (DAPL_DBG_TYPE_UTIL,
		      " ib_thread(%d,0x%x): ENTER: pipe %d at %d\n",
		      getpid(), g_ib_thread, 
		      g_ib_pipe[0], ib_at_get_fd());

 	/* Poll across pipe, CM, AT never changes */
	dapl_os_lock( &g_hca_lock );
	
	ufds[0].fd = g_ib_pipe[0];	/* pipe */
	ufds[0].events = POLLIN;
	ufds[1].fd = ib_at_get_fd();	/* uAT */
	ufds[1].events = POLLIN;
	
	while (!g_ib_destroy) {
		
		/* build ufds after pipe, at events */
		ufds[0].revents = 0;
		ufds[1].revents = 0;
		idx=1;

		/*  Walk HCA list and setup async and CQ events */
		if (!dapl_llist_is_empty(&g_hca_list))
			hca = dapl_llist_peek_head(&g_hca_list);
		else
			hca = NULL;
		
		while(hca) {
			ufds[++idx].fd = hca->ib_cm->fd; /* uCM */
			ufds[idx].events = POLLIN;
			ufds[idx].revents = 0;
			uhca[idx] = hca;
			ufds[++idx].fd = hca->ib_ctx->async_fd;	/* uASYNC */
			ufds[idx].events = POLLIN;
			ufds[idx].revents = 0;
			uhca[idx] = hca;
			
			if (hca->ib_cq != NULL) {
				ufds[++idx].fd = hca->ib_cq->fd; /* uCQ */
				ufds[idx].events = POLLIN;
				ufds[idx].revents = 0;
				uhca[idx] = hca;
			}
			
			hca = dapl_llist_next_entry(
				&g_hca_list,
				(DAPL_LLIST_ENTRY*)&hca->entry);
		}
		
		/* unlock, and setup poll */
		fds = idx+1;
		dapl_os_unlock(&g_hca_lock);
                ret = poll(ufds, fds, -1); 
		if (ret <= 0) {
			dapl_dbg_log(DAPL_DBG_TYPE_WARN,
				     " ib_thread(%d): ERR %s poll\n",
				     getpid(),strerror(errno));
                	dapl_os_lock(&g_hca_lock);
			continue;
		}

		/* check and process CM, CQ and ASYNC events, per device */
		for(idx=2;idx<fds;idx++) {
			if (ufds[idx].revents == POLLIN) {
				dapli_cq_event_cb(uhca[idx]);
				dapli_cm_event_cb(uhca[idx]);
				dapli_async_event_cb(uhca[idx]);
			}
		}
		
		/* check and process user events */
		if (ufds[0].revents == POLLIN) {

			read(g_ib_pipe[0], rbuf, 2);
			
			/* cleanup any device on list marked for destroy */
			for(idx=3;idx<fds;idx++) {
				if(uhca[idx] && uhca[idx]->destroy == 1) {
					dapl_os_lock(&g_hca_lock);
					dapl_llist_remove_entry(
						&g_hca_list, 
						(DAPL_LLIST_ENTRY*)
							&uhca[idx]->entry);
					dapl_os_unlock(&g_hca_lock);
					uhca[idx]->destroy = 2;
				}
			}
		}

		/* AT events */
		if (ufds[1].revents == POLLIN)
			dapli_at_event_cb();

		dapl_os_lock(&g_hca_lock);
	}
	dapl_dbg_log(DAPL_DBG_TYPE_UTIL," ib_thread(%d) EXIT\n",getpid());
	g_ib_destroy = 2;
	dapl_os_unlock(&g_hca_lock);	
}

